-module(eTpf).

-include("eTpf.hrl").

-export([
   trace/1
   , trace/2
   , trace/3
   , trace/4
   , stop/0
   , stop/1

]).


-spec trace(userInput()) -> ok.
trace(Input) ->
   trace(Input, tpTracerConsole).

-spec trace(userInput(), module()) -> ok.
trace(Input, TracerMod) ->
   trace(Input, TracerMod, undefined, #{}).

-spec trace(userInput(), module(), tracerOpts()) -> ok.
trace(Input, TracerMod, TracerOpts) ->
   trace(Input, TracerMod, TracerOpts, #{}).

-spec trace(userInput(), module(), tracerOpts(), traceOpts()) -> ok.
trace(Input, TracerMod, TracerOpts, TraceOpts) when is_list(Input) ->
   InputList = case is_list(Input) of true -> Input; _ -> [Input] end,
   do_trace(InputList, TracerMod, TracerOpts, TraceOpts).

do_trace(InputList, TracerMod, TracerOpts, TraceOpts) ->
   _ = application:ensure_all_started(eTpf),

   %% Start the pool of tracer processes.
   PoolID = maps:get(poolId, TraceOpts, ?eTpfPoolId),
   PoolSize = maps:get(poolSize, TraceOpts, erlang:system_info(schedulers)),
   true = PoolSize > 0,
   {ok, PoolPid} = supervisor:start_child(eTpf_sup, #{
      id => PoolID,
      start => {tpTracerPool, start_link, [PoolSize, TracerMod, TracerOpts]},
      restart => temporary,
      type => supervisor
   }),
   Tracers = tpTracerPool:tracers(PoolPid),
   TracersMap = maps:from_list(lists:zip(lists:seq(0, length(Tracers) - 1), Tracers)),
   Mode = maps:get(mode, TraceOpts, trace),
   Tem0InputList = flattenInput(InputList, []),
   Tem1InputList = ensurePattern(Tem0InputList),
   LastInputList = ensureScope(Tem1InputList),
   traceInput(LastInputList, #{mode => Mode, tracers => TracersMap}, TraceOpts),
   ok.

flattenInput([], Acc) ->
   lists:flatten(Acc);
flattenInput([{callback, Mod, Fun} | Tail], Acc) when is_atom(Mod), is_atom(Fun) ->
   Input = flattenInput(Mod:Fun(), []),
   flattenInput(Tail, [Input | Acc]);
flattenInput([{app, App} | Tail], Acc) when is_atom(App) ->
   _ = application:load(App),
   {ok, Mods} = application:get_key(App, modules),
   flattenInput(Tail, [Mods | Acc]);
flattenInput([Input | Tail], Acc) ->
   flattenInput(Tail, [Input | Acc]).

ensurePattern(Input) ->
   case [S || S = {scope, _} <- Input] of
      Input -> ['_' | Input];
      _ -> Input
   end.

ensureScope(Input) ->
   case [S || S = {scope, _} <- Input] of
      [] -> [{scope, [processes]} | Input];
      _ -> Input
   end.

traceInput([], _, _) ->
   ok;
traceInput([{scope, Scope} | Tail], TracerState, TraceOpts) ->
   %% We currently enable the following trace flags:
   %% - call: function calls
   %% - procs: process exit events; plus others we ignore
   %% - running: process being scheduled in/out
   %% - timestamp: events include the current timestamp
   %% - arity: function calls only include the arity, not arguments
   %% - return_to: return from functions
   %% - set_on_spawn: propagate trace flags to any children processes
   %%
   %% @todo It might be useful to count the number of sends
   %% or receives a function does.
   ExtraFlags = [running || maps:get(running, TraceOpts, false)] ++ [send || maps:get(send, TraceOpts, false)],
   [
      begin
         erlang:trace(PidPortSpec, true, [call, procs, monotonic_timestamp, arity, return_to, set_on_spawn, {tracer, tpTracerNif, TracerState} | ExtraFlags])
      end || PidPortSpec <- Scope
   ],
   traceInput(Tail, TracerState, TraceOpts);
traceInput([Mod | Tail], TracerState, TraceOpts) when is_atom(Mod) ->
   MatchSpec =
      case TraceOpts of
         #{process_dump := true} ->
            [{'_', [], [{message, {process_dump}}]}];
         _ ->
            true
      end,
   %% The module must be loaded before we attempt to trace it.
   _ = code:ensure_loaded(Mod),
   _ = erlang:trace_pattern({Mod, '_', '_'}, MatchSpec, [local]),
   traceInput(Tail, TracerState, TraceOpts).

stop() ->
   stop(?eTpfPoolId).

%% @todo Confirm that we don't need to stop tracing,
%% that just terminating the tracers is enough. The
%% NIF does cancel traces when tracers go away, but
%% better make sure.
stop(PoolID) ->
   supervisor:terminate_child(eTpf_sup, PoolID).
